"""
django style `lazy` object creation, the main bulk of processing
is turning {priceSize} into <PriceSize> objects but this is
pointless a lot of the time as the object is not used or can be
handled as a dict instead.

The inclusion of slots further reduces the processing time as well
as reducing memory.

This optimisation will improve normal streaming as well as
simulation, with more speed, less CPU + ram and minimal reduction
in usability.
"""


class SP:
    __slots__ = [
        "near_price",
        "far_price",
        "actual_sp",
        "back_stake_taken",
        "lay_liability_taken",
    ]

    def __init__(
        self,
        nearPrice: float = None,
        farPrice: float = None,
        backStakeTaken: list = None,
        layLiabilityTaken: list = None,
        actualSP: float = None,
    ):
        self.near_price = nearPrice
        self.far_price = farPrice
        self.actual_sp = actualSP
        self.back_stake_taken = backStakeTaken
        self.lay_liability_taken = layLiabilityTaken


class EX:
    __slots__ = ["available_to_back", "available_to_lay", "traded_volume"]

    def __init__(
        self,
        availableToBack: list = None,
        availableToLay: list = None,
        tradedVolume: list = None,
    ):
        self.available_to_back = availableToBack
        self.available_to_lay = availableToLay
        self.traded_volume = tradedVolume



import uuid
import unittest
import datetime

from flumine.strategy.runnercontext import RunnerContext


class RunnerContextTest(unittest.TestCase):
    def setUp(self):
        self.id_ = str(uuid.uuid1())
        self.context = RunnerContext(12345)

    def test_init(self):
        self.assertEqual(self.context.selection_id, 12345)
        self.assertFalse(self.context.invested)
        self.assertIsNone(self.context.datetime_last_placed)
        self.assertIsNone(self.context.datetime_last_reset)
        self.assertEqual(self.context.trades, [])
        self.assertEqual(self.context.live_trades, [])

    def test_place(self):
        self.context.place(self.id_)
        self.context.place(self.id_)  # append is ignored
        self.assertTrue(self.context.invested)
        self.assertIsNotNone(self.context.datetime_last_placed)
        self.assertEqual(self.context.trades, [self.id_])
        self.assertEqual(self.context.live_trades, [self.id_])

    def test_reset(self):
        self.context.live_trades = [self.id_]
        self.context.reset(self.id_)
        self.assertIsNotNone(self.context.datetime_last_reset)
        self.assertEqual(self.context.live_trades, [])

    def test_reset_error(self):
        self.context.reset(self.id_)
        self.assertIsNotNone(self.context.datetime_last_reset)
        self.assertEqual(self.context.live_trades, [])

    def test_executable_orders(self):
        self.assertFalse(self.context.executable_orders)
        self.context.live_trades = [self.id_]
        self.assertTrue(self.context.executable_orders)

    def test_trade_count(self):
        self.assertEqual(self.context.trade_count, 0)
        self.context.trades = [self.id_]
        self.assertEqual(self.context.trade_count, 1)

    def test_live_trade_count(self):
        self.assertEqual(self.context.live_trade_count, 0)
        self.context.live_trades = [self.id_]
        self.assertEqual(self.context.live_trade_count, 1)

    def test_placed_elapsed_seconds(self):
        self.assertIsNone(self.context.placed_elapsed_seconds)
        self.context.datetime_last_placed = datetime.datetime.utcnow()
        self.assertGreaterEqual(self.context.placed_elapsed_seconds, 0)

    def test_reset_elapsed_seconds(self):
        self.assertIsNone(self.context.reset_elapsed_seconds)
        self.context.datetime_last_reset = datetime.datetime.utcnow()
        self.assertGreaterEqual(self.context.reset_elapsed_seconds, 0)



import unittest
from unittest import mock

from flumine.strategy import strategy


class StrategiesTest(unittest.TestCase):
    def setUp(self) -> None:
        self.strategies = strategy.Strategies()

    def test_init(self):
        self.assertEqual(self.strategies._strategies, [])

    def test_call(self):
        mock_strategy = mock.Mock()
        mock_clients = mock.Mock()
        self.strategies(mock_strategy, mock_clients)
        self.assertEqual(self.strategies._strategies, [mock_strategy])
        mock_strategy.add.assert_called_with()
        self.assertEqual(mock_strategy.clients, mock_clients)

    def test_start(self):
        mock_strategy = mock.Mock()
        self.strategies._strategies.append(mock_strategy)
        self.strategies.start()
        mock_strategy.start.assert_called_with()

    def test_finish(self):
        mock_flumine = mock.Mock()
        mock_strategy = mock.Mock()
        self.strategies._strategies.append(mock_strategy)
        self.strategies.finish(mock_flumine)
        mock_strategy.finish.assert_called_with(mock_flumine)

    def test_iter(self):
        for i in self.strategies:
            assert i

    def test_len(self):
        self.assertEqual(len(self.strategies), 0)


class BaseStrategyTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_market_filter = mock.Mock()
        self.mock_market_data_filter = mock.Mock()
        self.streaming_timeout = 2
        self.conflate_ms = 100
        self.strategy = strategy.BaseStrategy(
            market_filter=self.mock_market_filter,
            market_data_filter=self.mock_market_data_filter,
            sports_data_filter=["cricketSubscription"],
            streaming_timeout=self.streaming_timeout,
            conflate_ms=self.conflate_ms,
            stream_class=strategy.MarketStream,
            name="test",
            context={"trigger": 0.123},
            max_selection_exposure=1,
            max_order_exposure=2,
            max_trade_count=3,
            max_live_trade_count=4,
            multi_order_trades=False,
        )

    def test_init(self):
        self.assertEqual(self.strategy.market_filter, self.mock_market_filter)
        self.assertEqual(self.strategy.market_data_filter, self.mock_market_data_filter)
        self.assertEqual(self.strategy.sports_data_filter, ["cricketSubscription"])
        self.assertEqual(self.strategy.streaming_timeout, self.streaming_timeout)
        self.assertEqual(self.strategy.conflate_ms, self.conflate_ms)
        self.assertEqual(self.strategy.stream_class, strategy.MarketStream)
        self.assertEqual(self.strategy._name, "test")
        self.assertEqual(self.strategy.context, {"trigger": 0.123})
        self.assertEqual(self.strategy.max_selection_exposure, 1)
        self.assertEqual(self.strategy.max_order_exposure, 2)
        self.assertIsNone(self.strategy.clients)
        self.assertEqual(self.strategy.max_trade_count, 3)
        self.assertEqual(self.strategy.max_live_trade_count, 4)
        self.assertEqual(self.strategy.streams, [])
        self.assertEqual(self.strategy.historic_stream_ids, [])
        self.assertEqual(self.strategy.name_hash, "a94a8fe5ccb19")
        self.assertFalse(self.strategy.multi_order_trades)
        self.assertEqual(strategy.STRATEGY_NAME_HASH_LENGTH, 13)
        self.assertEqual(
            strategy.DEFAULT_MARKET_DATA_FILTER,
            {
                "fields": [
                    "EX_ALL_OFFERS",
                    "EX_TRADED",
                    "EX_TRADED_VOL",
                    "EX_LTP",
                    "EX_MARKET_DEF",
                    "SP_TRADED",
                    "SP_PROJECTED",
                ]
            },
        )

    def test_check_market_no_subscribed(self):
        mock_market = mock.Mock()
        mock_market_book = mock.Mock()
        mock_market_book.streaming_unique_id = 12
        self.assertFalse(self.strategy.check_market(mock_market, mock_market_book))

    @mock.patch(
        "flumine.strategy.strategy.BaseStrategy.stream_ids",
        return_value=[12],
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "flumine.strategy.strategy.BaseStrategy.check_market_book", return_value=False
    )
    def test_check_market_check_fail(
        self, mock_check_market_book, mock_market_stream_ids
    ):
        mock_market = mock.Mock()
        mock_market_book = mock.Mock()
        mock_market_book.streaming_unique_id = 12
        self.assertFalse(self.strategy.check_market(mock_market, mock_market_book))
        mock_check_market_book.assert_called_with(mock_market, mock_market_book)
        mock_market_stream_ids.assert_called_with()

    @mock.patch(
        "flumine.strategy.strategy.BaseStrategy.stream_ids",
        return_value=[12],
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "flumine.strategy.strategy.BaseStrategy.check_market_book", return_value=True
    )
    def test_check_market_check_pass(
        self, mock_check_market_book, mock_market_stream_ids
    ):
        mock_market = mock.Mock()
        mock_market_book = mock.Mock()
        mock_market_book.streaming_unique_id = 12
        self.assertTrue(self.strategy.check_market(mock_market, mock_market_book))
        mock_check_market_book.assert_called_with(mock_market, mock_market_book)

    def test_add(self):
        self.strategy.add()

    def test_start(self):
        self.strategy.start()

    def test_check_market_book(self):
        self.assertFalse(self.strategy.check_market_book(None, None))

    def test_process_market_book(self):
        self.strategy.process_market_book(None, None)

    def test_process_sports_data(self):
        self.strategy.process_sports_data(None, None)

    def test_process_raw_data(self):
        self.strategy.process_raw_data(None, None, None)

    def test_process_orders(self):
        self.strategy.process_orders(None, None)

    def test_process_closed_market(self):
        self.strategy.process_closed_market(None, None)

    def test_finish(self):
        self.strategy.finish(mock.Mock())

    def test_remove_market(self):
        self.strategy._invested = {
            ("1.23", 456, 7): 1,
            ("1.23", 891, 7): 2,
            ("1.24", 112, 7): 3,
        }
        self.strategy.remove_market("1.23")
        self.assertEqual(self.strategy._invested, {("1.24", 112, 7): 3})

    def test_validate_order(self):
        mock_order = mock.Mock()
        runner_context = mock.Mock(
            trade_count=0,
            live_trade_count=0,
            placed_elapsed_seconds=None,
            reset_elapsed_seconds=None,
        )
        self.strategy.log_validation = True
        self.assertTrue(self.strategy.validate_order(runner_context, mock_order))
        # trade count
        runner_context.trade_count = 3
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # live trade count
        runner_context.trade_count = 1
        runner_context.live_trade_count = 4
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # place elapsed
        runner_context.trade_count = 1
        runner_context.live_trade_count = 1
        runner_context.placed_elapsed_seconds = 0.49
        mock_order.trade.place_reset_seconds = 0.5
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        # reset elapsed
        runner_context.trade_count = 1
        runner_context.live_trade_count = 1
        runner_context.placed_elapsed_seconds = None
        runner_context.reset_elapsed_seconds = 0.49
        mock_order.trade.reset_seconds = 0.5
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))

    def test_validate_order_multi(self):
        mock_order = mock.Mock()
        mock_order.trade.id = "test"
        runner_context = mock.Mock(
            trade_count=10,
            live_trade_count=10,
            live_trades=["test"],
            placed_elapsed_seconds=None,
            reset_elapsed_seconds=None,
        )
        self.assertFalse(self.strategy.validate_order(runner_context, mock_order))
        self.strategy.multi_order_trades = True
        self.assertTrue(self.strategy.validate_order(runner_context, mock_order))

    def test_executable_orders(self):
        mock_context = mock.Mock(executable_orders=True)
        self.strategy._invested = {("2", 456, 1): mock_context}
        self.assertFalse(self.strategy.has_executable_orders("1", 123, 1.0))
        self.assertFalse(self.strategy.has_executable_orders("2", 123, 1.0))
        self.assertTrue(self.strategy.has_executable_orders("2", 456, 1.0))

    def test_get_runner_context(self):
        mock_context = mock.Mock(invested=True)
        self.strategy._invested = {("2", 456, 0): mock_context}
        self.assertEqual(self.strategy.get_runner_context("2", 456, 0), mock_context)
        self.strategy.get_runner_context("2", 789, 0)
        self.assertEqual(len(self.strategy._invested), 2)

    def test_stream_ids(self):
        mock_stream = mock.Mock()
        mock_stream.stream_id = 321
        self.strategy.streams = [mock_stream]
        self.assertEqual(self.strategy.stream_ids, [321])
        self.strategy.historic_stream_ids = [123]
        self.assertEqual(self.strategy.stream_ids, [123])

    def test_info(self):
        self.assertEqual(
            self.strategy.info,
            {
                "conflate_ms": self.conflate_ms,
                "market_data_filter": self.mock_market_data_filter,
                "market_filter": self.mock_market_filter,
                "name": "test",
                "name_hash": "a94a8fe5ccb19",
                "stream_ids": [],
                "streaming_timeout": self.streaming_timeout,
                "context": {"trigger": 0.123},
                "max_live_trade_count": 4,
                "max_order_exposure": 2,
                "max_selection_exposure": 1,
                "max_trade_count": 3,
            },
        )

    def test_name(self):
        self.assertEqual(self.strategy.name, "test")

    def test_str(self):
        self.assertEqual(str(self.strategy), "test")



import unittest
from unittest import mock
from betfairlightweight.metadata import currency_parameters
from betfairlightweight.exceptions import BetfairError
from betconnect.exceptions import BetConnectException

from flumine.clients.clients import ExchangeType, Clients
from flumine.clients import BaseClient, BetfairClient, SimulatedClient, BetConnectClient
from flumine.clients import betfairclient
from flumine import exceptions


class ClientsTest(unittest.TestCase):
    def setUp(self):
        self.clients = Clients()

    def test_exchange_type(self):
        self.assertEqual(len(ExchangeType), 3)
        assert ExchangeType

    def test_init(self):
        self.assertEqual(self.clients._clients, [])
        self.assertEqual(
            self.clients._exchange_clients,
            {exchange: {} for exchange in ExchangeType},
        )

    def test_add_client(self):
        mock_client = mock.Mock(EXCHANGE=ExchangeType.BETFAIR, username="test")
        self.clients._clients = [mock_client]
        with self.assertRaises(exceptions.ClientError):
            self.clients.add_client(mock_client)

        with self.assertRaises(exceptions.ClientError):
            self.clients.add_client(mock.Mock(EXCHANGE="test", username="test"))

        self.clients._clients = []
        self.clients._exchange_clients[ExchangeType.BETFAIR][
            mock_client.username
        ] = mock_client
        with self.assertRaises(exceptions.ClientError):
            self.clients.add_client(mock_client)

        self.clients._clients = []
        self.clients._exchange_clients = {exchange: {} for exchange in ExchangeType}
        mock_client.EXCHANGE = ExchangeType.BETFAIR
        self.assertEqual(self.clients.add_client(mock_client), mock_client)
        self.assertEqual(self.clients._clients, [mock_client])
        self.assertEqual(
            self.clients._exchange_clients[mock_client.EXCHANGE],
            {mock_client.username: mock_client},
        )

    def test_get_default(self):
        self.clients._clients.append("howlandthehum")
        self.assertEqual(self.clients.get_default(), "howlandthehum")

    def test_get_betfair_default(self):
        mock_client_one = mock.Mock(EXCHANGE=ExchangeType.SIMULATED)
        mock_client_two = mock.Mock(EXCHANGE=ExchangeType.BETFAIR)
        self.clients._clients.append(mock_client_one)
        self.clients._clients.append(mock_client_two)
        self.assertEqual(self.clients.get_betfair_default(), mock_client_two)

    def test_get_client(self):
        self.clients._exchange_clients[ExchangeType.SIMULATED]["joejames"] = 12
        self.assertEqual(
            self.clients.get_client(ExchangeType.SIMULATED, "joejames"), 12
        )

    def test_login(self):
        mock_client = unittest.mock.Mock()
        self.clients._clients = [mock_client]
        self.clients.login()
        mock_client.login.assert_called_with()

    def test_keep_alive(self):
        mock_client = unittest.mock.Mock()
        self.clients._clients = [mock_client]
        self.clients.keep_alive()
        mock_client.keep_alive.assert_called_with()

    def test_logout(self):
        mock_client = unittest.mock.Mock()
        self.clients._clients = [mock_client]
        self.clients.logout()
        mock_client.logout.assert_called_with()

    def test_update_account_details(self):
        mock_client = unittest.mock.Mock()
        self.clients._clients = [mock_client]
        self.clients.update_account_details()
        mock_client.update_account_details.assert_called_with()

    def test_simulated(self):
        self.assertFalse(self.clients.simulated)
        self.clients._clients.append(mock.Mock(paper_trade=True))
        self.assertTrue(self.clients.simulated)

    def test_info(self):
        self.assertEqual(
            self.clients.info, {exchange.value: {} for exchange in ExchangeType}
        )
        mock_client = mock.Mock()
        self.clients._exchange_clients[ExchangeType.BETFAIR]["james"] = mock_client
        self.assertEqual(
            self.clients.info,
            {
                ExchangeType.BETFAIR.value: {"james": mock_client.info},
                ExchangeType.SIMULATED.value: {},
                ExchangeType.BETCONNECT.value: {},
            },
        )

    def test_iter(self):
        for i in self.clients:
            assert i

    def test_len(self):
        self.assertEqual(len(self.clients), 0)


class BaseClientTest(unittest.TestCase):
    def setUp(self):
        self.mock_betting_client = mock.Mock(lightweight=False)
        self.base_client = BaseClient(
            self.mock_betting_client, 1024, 100, 0.02, True, username="test"
        )

    def test_init(self):
        self.assertEqual(self.base_client.betting_client, self.mock_betting_client)
        self.assertEqual(self.base_client.transaction_limit, 1024)
        self.assertEqual(self.base_client.capital_base, 100)
        self.assertEqual(self.base_client.commission_base, 0.02)
        self.assertTrue(self.base_client.interactive_login)
        self.assertEqual(self.base_client._username, "test")
        self.assertIsNone(self.base_client.account_details)
        self.assertIsNone(self.base_client.account_funds)
        self.assertEqual(self.base_client.commission_paid, 0)
        self.assertIsNone(self.base_client.execution)
        self.assertEqual(self.base_client.trading_controls, [])
        self.assertTrue(self.base_client.order_stream)
        self.assertTrue(self.base_client.best_price_execution)
        self.assertTrue(self.base_client.min_bet_validation)
        self.assertFalse(self.base_client.paper_trade)

    def test_init_assert(self):
        with self.assertRaises(AssertionError):
            BaseClient(mock.Mock(lightweight=True), 1024, 100, 0.02, True)

    def test_login(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.login()

    def test_keep_alive(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.keep_alive()

    def test_logout(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.logout()

    def test_update_account_details(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.update_account_details()

    def test_add_execution(self):
        mock_flumine = mock.Mock()
        self.base_client.EXCHANGE = ExchangeType.SIMULATED
        self.base_client.add_execution(mock_flumine)
        self.assertEqual(self.base_client.execution, mock_flumine.simulated_execution)
        self.base_client.EXCHANGE = ExchangeType.BETFAIR
        self.base_client.add_execution(mock_flumine)
        self.assertEqual(self.base_client.execution, mock_flumine.betfair_execution)

    def test_add_execution_paper(self):
        self.base_client.paper_trade = True
        self.base_client.EXCHANGE = ExchangeType.BETFAIR
        mock_flumine = mock.Mock()
        self.base_client.add_execution(mock_flumine)
        self.assertEqual(self.base_client.execution, mock_flumine.simulated_execution)

    def test_add_transaction(self):
        mock_trading_control = mock.Mock()
        self.base_client.trading_controls.append(mock_trading_control)
        self.base_client.add_transaction(123, True)
        mock_trading_control.add_transaction.assert_called_with(123, True)

    def test_current_transaction_count_total(self):
        self.assertIsNone(self.base_client.current_transaction_count_total)
        mock_trading_control = mock.Mock(
            NAME="MAX_TRANSACTION_COUNT", current_transaction_count_total=123
        )
        self.base_client.trading_controls.append(mock_trading_control)
        self.assertEqual(self.base_client.current_transaction_count_total, 123)

    def test_transaction_count_total(self):
        self.assertIsNone(self.base_client.transaction_count_total)
        mock_trading_control = mock.Mock(
            NAME="MAX_TRANSACTION_COUNT", transaction_count_total=123
        )
        self.base_client.trading_controls.append(mock_trading_control)
        self.assertEqual(self.base_client.transaction_count_total, 123)

    def test_min_bet_size(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.min_bet_size

    def test_min_bet_payout(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.min_bet_payout

    def test_min_bsp_liability(self):
        with self.assertRaises(NotImplementedError):
            assert self.base_client.min_bsp_liability

    def test_username(self):
        self.assertEqual(
            self.base_client.username, self.base_client.betting_client.username
        )
        self.base_client.betting_client = None
        self.assertEqual(self.base_client.username, self.base_client._username)

    def test_info(self):
        self.assertTrue(self.base_client.info)


class BetfairClientTest(unittest.TestCase):
    def setUp(self):
        self.mock_betting_client = mock.Mock(lightweight=False)
        self.betfair_client = BetfairClient(self.mock_betting_client)

    def test_init(self):
        self.assertEqual(
            betfairclient.MIN_BET_SIZE, currency_parameters["GBP"]["min_bet_size"]
        )
        self.assertEqual(
            betfairclient.MIN_BSP_LIABILITY,
            currency_parameters["GBP"]["min_bsp_liability"],
        )
        self.assertEqual(
            betfairclient.MIN_BET_PAYOUT, currency_parameters["GBP"]["min_bet_payout"]
        )

    def test_login(self):
        self.betfair_client.login()
        self.mock_betting_client.login.assert_called_with()

    def test_login_no_certs(self):
        self.betfair_client.interactive_login = True
        self.betfair_client.login()
        self.mock_betting_client.login_interactive.assert_called_with()

    def test_login_error(self):
        self.betfair_client.betting_client.login.side_effect = BetfairError
        self.assertIsNone(self.betfair_client.login())
        self.mock_betting_client.login.assert_called_with()

    def test_keep_alive(self):
        self.mock_betting_client.session_expired = True
        self.betfair_client.keep_alive()
        self.mock_betting_client.keep_alive.assert_called_with()

    def test_keep_alive_error(self):
        self.betfair_client.betting_client.keep_alive.side_effect = BetfairError
        self.assertIsNone(self.betfair_client.keep_alive())
        self.mock_betting_client.keep_alive.assert_called_with()

    def test_logout(self):
        self.betfair_client.logout()
        self.mock_betting_client.logout.assert_called_with()

    def test_logout_error(self):
        self.betfair_client.betting_client.logout.side_effect = BetfairError
        self.assertIsNone(self.betfair_client.logout())
        self.mock_betting_client.logout.assert_called_with()

    @mock.patch("flumine.clients.betfairclient.BetfairClient._get_account_details")
    @mock.patch("flumine.clients.betfairclient.BetfairClient._get_account_funds")
    def test_update_account_details(self, mock_get_funds, mock_get_details):
        self.betfair_client.update_account_details()
        mock_get_funds.assert_called_with()
        mock_get_details.assert_called_with()
        self.assertEqual(self.betfair_client.account_details, mock_get_details())
        self.assertEqual(self.betfair_client.account_funds, mock_get_funds())

    def test__get_account_details(self):
        self.betfair_client._get_account_details()
        self.mock_betting_client.account.get_account_details.assert_called_with()

    def test__get_account_details_error(self):
        self.betfair_client.betting_client.account.get_account_details.side_effect = (
            BetfairError
        )
        self.assertIsNone(self.betfair_client._get_account_details())
        self.mock_betting_client.account.get_account_details.assert_called_with()

    def test__get_account_funds(self):
        self.betfair_client._get_account_funds()
        self.mock_betting_client.account.get_account_funds.assert_called_with()

    def test__get_account_funds_error(self):
        self.betfair_client.betting_client.account.get_account_funds.side_effect = (
            BetfairError
        )
        self.assertIsNone(self.betfair_client._get_account_funds())
        self.mock_betting_client.account.get_account_funds.assert_called_with()

    def test_min_bet_size(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = "GBP"
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_size, 1)

    def test_min_bet_size_none(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_size, 1)

    def test_min_bet_size_ac_none(self):
        mock_account_details = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_size, 1)

    def test_min_bsp_liability(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = "USD"
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bsp_liability, 20)

    def test_min_bsp_liability_none(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bsp_liability, 10)

    def test_min_bsp_liability_ac_none(self):
        mock_account_details = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bsp_liability, 10)

    def test_min_bet_payout(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = "GBP"
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_payout, 10)

    def test_min_bet_payout_none(self):
        mock_account_details = mock.Mock()
        mock_account_details.currency_code = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_payout, 10)

    def test_min_bet_payout_ac_none(self):
        mock_account_details = None
        self.betfair_client.account_details = mock_account_details
        self.assertEqual(self.betfair_client.min_bet_payout, 10)


class SimulatedClientTest(unittest.TestCase):
    def setUp(self):
        self.simulated_client = SimulatedClient()

    def test_login(self):
        self.simulated_client.login()

    def test_keep_alive(self):
        self.simulated_client.keep_alive()

    def test_logout(self):
        self.simulated_client.logout()

    @mock.patch("flumine.clients.simulatedclient.AccountDetails")
    def test_update_account_details(self, mock_account_details):
        self.simulated_client.update_account_details()
        self.assertEqual(self.simulated_client.account_details, mock_account_details())

    def test_min_bet_size(self):
        self.simulated_client.update_account_details()
        self.assertEqual(self.simulated_client.min_bet_size, 1)

    def test_min_bsp_liability(self):
        self.simulated_client.update_account_details()
        self.assertEqual(self.simulated_client.min_bsp_liability, 10)

    def test_min_bet_payout(self):
        self.simulated_client.update_account_details()
        self.assertEqual(self.simulated_client.min_bet_payout, 10)

    def test_username(self):
        self.assertEqual(
            self.simulated_client.username,
            self.simulated_client._username,
        )


class BetConnectClientTest(unittest.TestCase):
    def setUp(self):
        self.mock_betting_client = mock.Mock()
        del self.mock_betting_client.lightweight
        self.betconnect_client = BetConnectClient(self.mock_betting_client)

    def test_login(self):
        self.betconnect_client.login()
        self.mock_betting_client.account.login.assert_called_with()

    def test_login_error(self):
        self.betconnect_client.betting_client.account.login.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client.login())
        self.mock_betting_client.account.login.assert_called_with()

    def test_keep_alive(self):
        self.mock_betting_client.session_expired = True
        self.betconnect_client.keep_alive()
        self.mock_betting_client.account.refresh_session_token.assert_called_with()

    def test_keep_alive_error(self):
        self.betconnect_client.betting_client.account.refresh_session_token.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client.keep_alive())
        self.mock_betting_client.account.refresh_session_token.assert_called_with()

    def test_logout(self):
        self.betconnect_client.logout()
        self.mock_betting_client.account.logout.assert_called_with()

    def test_logout_error(self):
        self.betconnect_client.betting_client.account.logout.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client.logout())
        self.mock_betting_client.account.logout.assert_called_with()

    @mock.patch(
        "flumine.clients.betconnectclient.BetConnectClient._get_account_details"
    )
    @mock.patch("flumine.clients.betconnectclient.BetConnectClient._get_account_funds")
    def test_update_account_details(self, mock_get_funds, mock_get_details):
        self.betconnect_client.update_account_details()
        mock_get_funds.assert_called_with()
        mock_get_details.assert_called_with()
        self.assertEqual(self.betconnect_client.account_details, mock_get_details())
        self.assertEqual(self.betconnect_client.account_funds, mock_get_funds())

    def test__get_account_details(self):
        self.betconnect_client._get_account_details()
        self.mock_betting_client.account.get_user_preferences.assert_called_with()

    def test__get_account_details_error(self):
        self.betconnect_client.betting_client.account.get_user_preferences.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client._get_account_details())
        self.mock_betting_client.account.get_user_preferences.assert_called_with()

    def test__get_account_funds(self):
        self.betconnect_client._get_account_funds()
        self.mock_betting_client.account.get_balance.assert_called_with()

    def test__get_account_funds_error(self):
        self.betconnect_client.betting_client.account.get_balance.side_effect = (
            BetConnectException
        )
        self.assertIsNone(self.betconnect_client._get_account_funds())
        self.mock_betting_client.account.get_balance.assert_called_with()



import uuid
import time
from enum import Enum
from typing import Iterator, Optional
from betfairlightweight.metadata import order_limits

from ..events.events import BaseEvent, EventType, QueueType
from ..clients.clients import ExchangeType
from .. import config
from .order import BaseOrder, OrderStatus


class OrderPackageType(Enum):
    PLACE = "Place"
    CANCEL = "Cancel"
    REPLACE = "Replace"
    UPDATE = "Update"


class BaseOrderPackage(BaseEvent):
    """
    Data structure for multiple orders,
    temporary to allow execution
    """

    EVENT_TYPE = EventType.ORDER_PACKAGE
    QUEUE_TYPE = QueueType.HANDLER
    EXCHANGE = None

    def __init__(
        self,
        client,
        market_id: str,
        orders: list,
        package_type: OrderPackageType,
        bet_delay: int,
        async_: bool = False,
        market_version: int = None,
    ):
        super(BaseOrderPackage, self).__init__(None)
        self.id = uuid.uuid1()
        self.client = client
        self.market_id = market_id
        self._orders = orders
        self.package_type = package_type
        self.async_ = async_
        self._market_version = market_version
        self.customer_strategy_ref = config.customer_strategy_ref
        self._retry = True
        self._max_retries = 3  # will retry 3 times
        self._retry_count = 0
        # following used for simulated execution
        self.processed = False
        self.bet_delay = bet_delay
        self.simulated_delay = self.calc_simulated_delay()

    def retry(self):
        if self._retry and self._retry_count < self._max_retries:
            time.sleep(self._retry_count)  # back-off
            self._retry_count += 1
            return True
        return False

    def reset_orders(self, complete: bool = False) -> None:
        for order in self:
            with order.trade:
                if complete:
                    order.execution_complete()
                else:
                    order.executable()

    def calc_simulated_delay(self) -> float:
        if self.client.execution.EXCHANGE == ExchangeType.SIMULATED:
            if self.package_type == OrderPackageType.PLACE:
                return config.place_latency + self.bet_delay
            elif self.package_type == OrderPackageType.CANCEL:
                return config.cancel_latency
            elif self.package_type == OrderPackageType.UPDATE:
                return config.update_latency
            elif self.package_type == OrderPackageType.REPLACE:
                return config.replace_latency + self.bet_delay

    @property
    def place_instructions(self) -> dict:
        raise NotImplementedError

    @property
    def cancel_instructions(self) -> dict:
        raise NotImplementedError

    @property
    def update_instructions(self) -> dict:
        raise NotImplementedError

    @property
    def replace_instructions(self) -> dict:
        raise NotImplementedError

    @classmethod
    def order_limit(cls, package_type: OrderPackageType) -> int:
        raise NotImplementedError

    @property
    def orders(self) -> list:
        return [o for o in self._orders if o.status != OrderStatus.VIOLATION]

    @property
    def retry_count(self) -> int:
        return self._retry_count

    @property
    def info(self) -> dict:
        return {
            "id": self.id,
            "client": self.client,
            "market_id": self.market_id,
            "orders": [o.id for o in self._orders],
            "order_count": len(self),
            "package_type": self.package_type.value,
            "customer_strategy_ref": self.customer_strategy_ref,
            "bet_delay": self.bet_delay,
            "market_version": self._market_version,
            "retry": self._retry,
            "retry_count": self._retry_count,
            "async": self.async_,
        }

    @property
    def market_version(self) -> Optional[dict]:
        if self._market_version:
            return {"version": self._market_version}

    def __iter__(self) -> Iterator[BaseOrder]:
        return iter(self.orders)

    def __len__(self) -> int:
        return len(self.orders)


class BetfairOrderPackage(BaseOrderPackage):

    EXCHANGE = ExchangeType.BETFAIR

    @property
    def place_instructions(self):
        return [order.create_place_instruction() for order in self]

    @property
    def cancel_instructions(self):
        return [
            order.create_cancel_instruction() for order in self
        ]  # todo? if order.size_remaining > 0

    @property
    def update_instructions(self):
        return [order.create_update_instruction() for order in self]

    @property
    def replace_instructions(self):
        return [
            order.create_replace_instruction()
            for order in self
            if order.status != OrderStatus.EXECUTION_COMPLETE
        ]

    @classmethod
    def order_limit(cls, package_type: OrderPackageType) -> int:
        if package_type == OrderPackageType.PLACE:
            return order_limits["placeOrders"]
        elif package_type == OrderPackageType.CANCEL:
            return order_limits["cancelOrders"]
        elif package_type == OrderPackageType.UPDATE:
            return order_limits["updateOrders"]
        elif package_type == OrderPackageType.REPLACE:
            return order_limits["replaceOrders"]



import logging
from collections import defaultdict

from .utils import SimulatedDateTime
from ..baseflumine import BaseFlumine
from ..clients import BaseClient
from ..events import events
from .. import utils
from ..exceptions import RunError
from ..order.order import OrderTypes

logger = logging.getLogger(__name__)


class FlumineSimulation(BaseFlumine):
    """
    Single threaded implementation of flumine
    for simulating strategies with betfair
    historic (or self recorded) streaming data.
    """

    SIMULATED = True

    def __init__(self, client: BaseClient = None):
        super(FlumineSimulation, self).__init__(client)
        self.simulated_datetime = SimulatedDateTime()
        self.handler_queue = []

    def run(self) -> None:
        if not self.clients.simulated:
            raise RunError(
                "Incorrect client provided, only a Simulated client can be used when simulating"
            )
        with self:
            with self.simulated_datetime:
                """
                list of either single stream or complete events depending
                on event_processing flag:
                   single: {None: [<Stream 1>], [<Stream 2>, ..]}
                   event: {123: [<Stream 1>, <Stream 2>, ..], 456: [..]}
                Event data to be muxed/processed chronologically as per
                live rather than single which is per market in isolation.
                """
                event_streams = defaultdict(list)  # eventId: [<Stream>, ..]
                for stream in self.streams:
                    event_id = stream.event_id if stream.event_processing else None
                    event_streams[event_id].append(stream)

                for event_id, streams in event_streams.items():
                    if event_id and len(streams) > 1:
                        logger.info(
                            "Starting historical event '{0}'".format(event_id),
                            extra={
                                "event_id": event_id,
                                "markets": [s.market_filter for s in streams],
                            },
                        )
                        self.simulated_datetime.reset_real_datetime()
                        # create cycles
                        cycles = []  # [[epoch, [MarketBook], gen], ..]
                        for stream in streams:
                            stream_gen = stream.create_generator()()
                            market_book = next(stream_gen)
                            publish_time_epoch = market_book[0].publish_time_epoch
                            cycles.append([publish_time_epoch, market_book, stream_gen])
                        # process cycles
                        while cycles:
                            # order by epoch
                            cycles.sort(key=lambda x: x[0])
                            # get current
                            _, market_book, stream_gen = cycles.pop(0)
                            # process current
                            self._process_market_books(
                                events.MarketBookEvent(market_book)
                            )
                            # gen next
                            try:
                                market_book = next(stream_gen)
                            except StopIteration:
                                continue
                            publish_time_epoch = market_book[0].publish_time_epoch
                            # add back
                            cycles.append([publish_time_epoch, market_book, stream_gen])
                        self.handler_queue.clear()
                        logger.info("Completed historical event '{0}'".format(event_id))
                    else:
                        for stream in streams:
                            logger.info(
                                "Starting historical market '{0}'".format(
                                    stream.market_filter
                                ),
                                extra={"market": stream.market_filter},
                            )
                            self.simulated_datetime.reset_real_datetime()
                            stream_gen = stream.create_generator()
                            for event in stream_gen():
                                self._process_market_books(
                                    events.MarketBookEvent(event)
                                )
                            self.handler_queue.clear()
                            logger.info(
                                "Completed historical market '{0}'".format(
                                    stream.market_filter
                                )
                            )
                self._process_end_flumine()
                logger.info("Simulation complete")

    def _process_market_books(self, event: events.MarketBookEvent) -> None:
        # todo DRY!
        for market_book in event.event:
            market_id = market_book.market_id
            self.simulated_datetime(market_book.publish_time)

            # check if there are orders to process (limited to current market only)
            if self.handler_queue:
                self._check_pending_packages(market_id)

            if market_book.status == "CLOSED":
                self._process_close_market(event=events.CloseMarketEvent(market_book))
                continue

            # get market
            market = self.markets.markets.get(market_id)
            if market is None:
                market = self._add_market(market_id, market_book)
                self.log_control(events.MarketEvent(market))
            elif market.closed:
                self.markets.add_market(market_id, market)

            # process market
            market(market_book)

            # process middleware
            for middleware in self._market_middleware:
                utils.call_middleware_error_handling(middleware, market)

            # process current orders
            if market.blotter.active:
                self._process_simulated_orders(market)

            for strategy in self.strategies:
                if utils.call_strategy_error_handling(
                    strategy.check_market, market, market_book
                ):
                    utils.call_strategy_error_handling(
                        strategy.process_market_book, market, market_book
                    )

    def process_order_package(self, order_package) -> None:
        # place in pending list (wait for latency+delay)
        self.handler_queue.append(order_package)

    def _process_simulated_orders(self, market) -> None:
        """Remove order from blotter live
        orders if complete and process
        orders through strategies
        """
        blotter = market.blotter
        for order in blotter.live_orders:
            if order.complete:
                blotter.complete_order(order)
            else:
                if order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
                    if order.size_remaining == 0:
                        order.execution_complete()
                        blotter.complete_order(order)
                elif order.order_type.ORDER_TYPE in [
                    OrderTypes.LIMIT_ON_CLOSE,
                    OrderTypes.MARKET_ON_CLOSE,
                ]:
                    if order.current_order.status == "EXECUTION_COMPLETE":
                        order.execution_complete()
                        blotter.complete_order(order)
        for strategy in self.strategies:
            strategy_orders = blotter.strategy_orders(strategy)
            if strategy_orders:
                utils.call_process_orders_error_handling(
                    strategy, market, strategy_orders
                )

    def _check_pending_packages(self, market_id: str) -> None:
        processed = []
        for order_package in self.handler_queue:
            if (
                order_package.market_id == market_id
                and order_package.elapsed_seconds > order_package.simulated_delay
            ):
                order_package.client.execution.handler(order_package)
                processed.append(order_package)
        for p in processed:
            self.handler_queue.remove(p)

    def __repr__(self) -> str:
        return "<FlumineSimulation>"

    def __str__(self) -> str:
        return "<FlumineSimulation>"



import logging
from typing import Type, Iterator, Union, List
from betfairlightweight import filters
from betfairlightweight.resources import MarketBook, Race, CricketMatch

from .runnercontext import RunnerContext
from ..markets.market import Market
from ..streams.marketstream import BaseStream, MarketStream
from ..utils import create_cheap_hash, STRATEGY_NAME_HASH_LENGTH

logger = logging.getLogger(__name__)

DEFAULT_MARKET_DATA_FILTER = filters.streaming_market_data_filter(
    fields=[
        "EX_ALL_OFFERS",
        "EX_TRADED",
        "EX_TRADED_VOL",
        "EX_LTP",
        "EX_MARKET_DEF",
        "SP_TRADED",
        "SP_PROJECTED",
    ]
)


class BaseStrategy:

    """
    Strategy object to process MarketBook data
    from streams, order placement and handling
    logic to be added where required. Only
    MarketBooks from provided filter and data
    filter are processed.
    Runner context available to store current
    live trades.
    """

    def __init__(
        self,
        market_filter: Union[dict, list],
        market_data_filter: dict = None,
        sports_data_filter: List[
            str
        ] = None,  # 'raceSubscription', 'cricketSubscription'
        streaming_timeout: float = None,
        conflate_ms: int = None,
        stream_class: Type[BaseStream] = MarketStream,
        name: str = None,
        context: dict = None,
        max_selection_exposure: float = 100,
        max_order_exposure: float = 10,
        max_trade_count: int = 1e6,
        max_live_trade_count: int = 1,
        multi_order_trades: bool = False,
    ):
        """
        :param market_filter: Streaming market filter dict or list of market filters
        :param market_data_filter: Streaming market data filter
        :param sports_data_filter: Streaming sports data filter (e.g. ["raceSubscription"])
        :param streaming_timeout: Streaming timeout in seconds, will call snap() on cache
        :param conflate_ms: Streaming conflation
        :param stream_class: Can be Market or Data (raw)
        :param name: Strategy name (will default to class name)
        :param context: Dictionary holding additional user specific vars
        :param max_selection_exposure: Max exposure per selection
        :param max_order_exposure: Max exposure per order
        :param max_trade_count: max total number of trades per runner
        :param max_live_trade_count: max live (with executable orders) trades per runner
        :param multi_order_trades: allow multiple live orders per trade
        """
        self.market_filter = market_filter
        self.market_data_filter = market_data_filter or DEFAULT_MARKET_DATA_FILTER
        self.sports_data_filter = sports_data_filter or []
        self.streaming_timeout = streaming_timeout
        self.conflate_ms = conflate_ms
        self.stream_class = stream_class
        self._name = name
        self.context = context or {}
        self.max_selection_exposure = max_selection_exposure
        self.max_order_exposure = max_order_exposure
        self.clients = None
        self.max_trade_count = max_trade_count
        self.max_live_trade_count = max_live_trade_count
        self.multi_order_trades = multi_order_trades

        self._invested = {}  # {(marketId, selectionId, handicap): RunnerContext}
        self.streams = []  # list of streams strategy is subscribed
        self.historic_stream_ids = []
        # cache
        self.name_hash = create_cheap_hash(self.name, STRATEGY_NAME_HASH_LENGTH)

    def check_market(self, market: Market, market_book: MarketBook) -> bool:
        if market_book.streaming_unique_id not in self.stream_ids:
            return False  # strategy not subscribed to market stream
        elif self.check_market_book(market, market_book):
            return True
        else:
            return False

    def add(self) -> None:
        # called when strategy is added to framework
        return

    def start(self) -> None:
        # called when flumine starts but before streams start
        # e.g. subscribe to extra streams
        return

    def check_market_book(self, market: Market, market_book: MarketBook) -> bool:
        # process_market_book only executed if this returns True
        return False

    def process_market_book(self, market: Market, market_book: MarketBook) -> None:
        # process marketBook; place/cancel/replace orders
        return

    def process_sports_data(
        self, market: Market, sports_data: Union[Race, CricketMatch]
    ) -> None:
        # process sports data
        return

    def process_raw_data(self, clk: str, publish_time: int, datum: dict) -> None:
        return

    def process_orders(self, market: Market, orders: list) -> None:
        # process list of Order objects for strategy and Market
        return

    def process_closed_market(self, market: Market, market_book: MarketBook) -> None:
        # process marketBook after closure
        return

    def finish(self, flumine) -> None:
        # called before flumine ends
        return

    def remove_market(self, market_id: str) -> None:
        to_remove = []
        for invested in self._invested:
            if invested[0] == market_id:
                to_remove.append(invested)
        for i in to_remove:
            del self._invested[i]

    def validate_order(self, runner_context: RunnerContext, order) -> bool:
        # allow multiple orders per trade
        if self.multi_order_trades:
            if order.trade.id in runner_context.live_trades:
                return True
        # validate context
        reset_elapsed_seconds = runner_context.reset_elapsed_seconds
        if reset_elapsed_seconds and reset_elapsed_seconds < order.trade.reset_seconds:
            order.violation_msg = "strategy.validate_order failed: reset_elapsed_seconds (%s) < reset_seconds (%s)".format(
                reset_elapsed_seconds,
                order.trade.reset_seconds,
            )
            return False

        placed_elapsed_seconds = runner_context.placed_elapsed_seconds
        if (
            placed_elapsed_seconds
            and placed_elapsed_seconds < order.trade.place_reset_seconds
        ):
            order.violation_msg = "strategy.validate_order failed: placed_elapsed_seconds (%s) < place_reset_seconds (%s)".format(
                placed_elapsed_seconds,
                order.trade.place_reset_seconds,
            )
            return False

        if runner_context.trade_count >= self.max_trade_count:
            order.violation_msg = "strategy.validate_order failed: trade_count (%s) >= max_trade_count (%s)".format(
                runner_context.trade_count, self.max_trade_count
            )
            return False
        elif runner_context.live_trade_count >= self.max_live_trade_count:
            order.violation_msg = "strategy.validate_order failed: live_trade_count (%s) >= max_live_trade_count (%s)".format(
                runner_context.live_trade_count, self.max_live_trade_count
            )
            return False

        return True

    def has_executable_orders(
        self, market_id: str, selection_id: int, handicap: float = 0
    ) -> bool:
        runner_context = self.get_runner_context(market_id, selection_id, handicap)
        return runner_context.executable_orders

    def get_runner_context(
        self, market_id: str, selection_id: int, handicap: float = 0
    ) -> RunnerContext:
        try:
            return self._invested[(market_id, selection_id, handicap)]
        except KeyError:
            self._invested[
                (market_id, selection_id, handicap)
            ] = runner_context = RunnerContext(selection_id)
            return runner_context

    @property
    def stream_ids(self) -> list:
        if self.historic_stream_ids:
            return self.historic_stream_ids
        else:
            return [stream.stream_id for stream in self.streams]

    @property
    def info(self) -> dict:
        return {
            "name": self.name,
            "market_filter": self.market_filter,
            "market_data_filter": self.market_data_filter,
            "streaming_timeout": self.streaming_timeout,
            "conflate_ms": self.conflate_ms,
            "stream_ids": self.stream_ids,
            "max_selection_exposure": self.max_selection_exposure,
            "max_order_exposure": self.max_order_exposure,
            "max_live_trade_count": self.max_live_trade_count,
            "max_trade_count": self.max_trade_count,
            "context": self.context,
            "name_hash": self.name_hash,
        }

    @property
    def name(self) -> str:
        return self._name or self.__class__.__name__

    def __str__(self):
        return "{0}".format(self.name)


class Strategies:
    def __init__(self):
        self._strategies = []

    def __call__(self, strategy: BaseStrategy, clients) -> None:
        if strategy.name in [s.name for s in self]:
            logger.warning("Strategy of same name '{0}' already added".format(strategy))
        strategy.clients = clients
        self._strategies.append(strategy)
        strategy.add()

    def start(self) -> None:
        for s in self:
            s.start()

    def finish(self, flumine) -> None:
        for s in self:
            s.finish(flumine)

    @property
    def hashes(self) -> dict:
        return {strategy.name_hash: strategy for strategy in self}

    def __iter__(self) -> Iterator[BaseStrategy]:
        return iter(self._strategies)

    def __len__(self) -> int:
        return len(self._strategies)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "requirements.txt")) as f:
    INSTALL_REQUIRES = f.read().splitlines()

about = {}
with open(os.path.join(here, "flumine", "__version__.py"), "r") as f:
    exec(f.read(), about)

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name=about["__title__"],
    version=about["__version__"],
    packages=find_packages(
        exclude=[
            "examples",
            "examples.strategies",
            "examples.workers",
            "examples.middleware",
            "examples.controls",
        ]
    ),
    package_dir={"flumine": "flumine"},
    install_requires=INSTALL_REQUIRES,
    url=about["__url__"],
    license=about["__license__"],
    author=about["__author__"],
    author_email="a@unknown.com",
    description=about["__description__"],
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="tests",
)
